@@ -29,6 +29,9 @@
 #include <config.h>
 #include <stdlib.h>
 
+#define _WIN32_WINNT 0x0500
+#define WINVER _WIN32_WINNT
+
 #include "gdk.h" /* gdk_rectangle_intersect */
 #include "gdkevents.h"
 #include "gdkpixmap.h"
@@ -37,6 +40,35 @@
 #include "gdkprivate-win32.h"
 #include "gdkinput-win32.h"
 
+#if defined(_MSC_VER) && (WINVER < 0x0500)
+#define GetAncestor(hwnd,what) _gdk_win32_get_ancestor_parent(hwnd)
+
+static HWND
+_gdk_win32_get_ancestor_parent (HWND hwnd)
+{
+#ifndef GA_PARENT
+#  define GA_PARENT 1 
+#endif
+  typedef HWND (WINAPI *PFN_GetAncestor) (HWND,UINT);
+  static PFN_GetAncestor p_GetAncestor = NULL;
+  static gboolean once = FALSE;
+  
+  if (!once)
+    {
+      HMODULE user32;
+
+      user32 = GetModuleHandle ("user32.dll");
+      p_GetAncestor = (PFN_GetAncestor)GetProcAddress (user32, "GetAncestor");
+      once = TRUE;
+    }
+  if (p_GetAncestor)
+    return p_GetAncestor (hwnd, GA_PARENT);
+  else /* not completely right, but better than nothing ? */
+    return GetParent (hwnd);
+}
+
+#endif
+
 #if 0
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include <stdio.h>
@@ -55,6 +87,10 @@ static void gdk_window_impl_win32_finalize   (GObject                 *object);
 
 static gpointer parent_class = NULL;
 
+#define WINDOW_IS_TOPLEVEL(window)		   \
+  (GDK_WINDOW_TYPE (window) != GDK_WINDOW_CHILD && \
+   GDK_WINDOW_TYPE (window) != GDK_WINDOW_FOREIGN)
+
 GType
 _gdk_window_impl_win32_get_type (void)
 {
@@ -94,7 +130,7 @@ gdk_window_impl_win32_init (GdkWindowImplWin32 *impl)
 {
   impl->width = 1;
   impl->height = 1;
-
+  impl->toplevel_window_type = -1;
   impl->hcursor = NULL;
   impl->hicon_big = NULL;
   impl->hicon_small = NULL;
@@ -278,7 +314,7 @@ _gdk_root_window_size_init (void)
   GdkRectangle rect;
   int i;
 
-  impl = GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) _gdk_parent_root)->impl);
+  impl = GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) _gdk_root)->impl);
   rect = _gdk_monitors[0];
   for (i = 1; i < _gdk_num_monitors; i++)
     gdk_rectangle_union (&rect, _gdk_monitors+i, &rect);
@@ -293,13 +329,13 @@ _gdk_windowing_window_init (void)
   GdkWindowObject *private;
   GdkDrawableImplWin32 *draw_impl;
 
-  g_assert (_gdk_parent_root == NULL);
+  g_assert (_gdk_root == NULL);
   
-  _gdk_parent_root = g_object_new (GDK_TYPE_WINDOW, NULL);
-  private = (GdkWindowObject *)_gdk_parent_root;
+  _gdk_root = g_object_new (GDK_TYPE_WINDOW, NULL);
+  private = (GdkWindowObject *)_gdk_root;
   draw_impl = GDK_DRAWABLE_IMPL_WIN32 (private->impl);
   
-  draw_impl->handle = _gdk_root_window;
+  draw_impl->handle = GetDesktopWindow ();
   draw_impl->wrapper = GDK_DRAWABLE (private);
   draw_impl->colormap = gdk_colormap_get_system ();
   g_object_ref (draw_impl->colormap);
@@ -311,9 +347,9 @@ _gdk_windowing_window_init (void)
  
   _gdk_window_init_position (GDK_WINDOW (private));
 
-  gdk_win32_handle_table_insert (&_gdk_root_window, _gdk_parent_root);
+  gdk_win32_handle_table_insert ((HANDLE *) &draw_impl->handle, _gdk_root);
 
-  GDK_NOTE (MISC, g_print ("_gdk_parent_root=%p\n", GDK_WINDOW_HWND (_gdk_parent_root)));
+  GDK_NOTE (MISC, g_print ("_gdk_root=%p\n", GDK_WINDOW_HWND (_gdk_root)));
 }
 
 static const gchar *
@@ -453,11 +489,13 @@ gdk_window_new_internal (GdkWindow     *parent,
 			 gint           attributes_mask,
 			 gboolean       from_set_skip_taskbar_hint)
 {
+  HWND hwndNew;
   HANDLE hparent;
   ATOM klass = 0;
   DWORD dwStyle = 0, dwExStyle;
   RECT rect;
   GdkWindow *window;
+  GdkWindow *orig_parent;
   GdkWindowObject *private;
   GdkWindowImplWin32 *impl;
   GdkDrawableImplWin32 *draw_impl;
@@ -473,13 +511,15 @@ gdk_window_new_internal (GdkWindow     *parent,
   if (!parent)
     {
       screen = gdk_screen_get_default ();
-      parent = gdk_screen_get_root_window (screen);
+      parent = _gdk_root;
     }
   else
     screen = gdk_drawable_get_screen (parent);
 
   g_return_val_if_fail (GDK_IS_WINDOW (parent), NULL);
   
+  orig_parent = parent;
+
   GDK_NOTE (MISC,
 	    g_print ("gdk_window_new: %s\n",
 		     (attributes->window_type == GDK_WINDOW_TOPLEVEL ? "TOPLEVEL" :
@@ -503,7 +543,7 @@ gdk_window_new_internal (GdkWindow     *parent,
    * of the root window, except for actual creation.
    */
   if (GDK_WINDOW_TYPE (parent) == GDK_WINDOW_FOREIGN)
-    parent = _gdk_parent_root;
+    parent = _gdk_root;
   
   private->parent = (GdkWindowObject *)parent;
 
@@ -536,7 +576,7 @@ gdk_window_new_internal (GdkWindow     *parent,
        * attributes->window_type for input-only windows
        * before
        */
-      if (GDK_WINDOW_TYPE (parent) == GDK_WINDOW_ROOT)
+      if (parent == _gdk_root)
 	private->window_type = GDK_WINDOW_TEMP;
       else
 	private->window_type = GDK_WINDOW_CHILD;
@@ -569,36 +609,44 @@ gdk_window_new_internal (GdkWindow     *parent,
       private->input_only = TRUE;
       draw_impl->colormap = gdk_screen_get_system_colormap (screen);
       g_object_ref (draw_impl->colormap);
-      GDK_NOTE (MISC, g_print ("...GDK_INPUT_ONLY, system colormap"));
+      GDK_NOTE (MISC, g_print ("... GDK_INPUT_ONLY, system colormap"));
     }
 
   switch (private->window_type)
     {
     case GDK_WINDOW_TOPLEVEL:
-      dwStyle = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN;
-      hparent = _gdk_root_window;
-      offset_x = _gdk_offset_x;
-      offset_y = _gdk_offset_y;
+    case GDK_WINDOW_DIALOG:
+      if (parent != _gdk_root)
+	{
+	  g_warning (G_STRLOC ": Toplevel windows must be created as children\n"
+		     "of a window of type GDK_WINDOW_ROOT or GDK_WINDOW_FOREIGN");
+	  hparent = GetDesktopWindow ();
+	}
+      /* Children of foreign windows aren't toplevel windows */
+      if (GDK_WINDOW_TYPE (orig_parent) == GDK_WINDOW_FOREIGN)
+	{
+	  dwStyle = WS_CHILDWINDOW | WS_CLIPCHILDREN;
+	}
+      else
+	{
+	  if (private->window_type == GDK_WINDOW_TOPLEVEL)
+	    dwStyle = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN;
+	  else
+	    dwStyle = WS_OVERLAPPED | WS_MINIMIZEBOX | WS_SYSMENU | WS_CAPTION | WS_THICKFRAME | WS_CLIPCHILDREN;
+
+	  offset_x = _gdk_offset_x;
+	  offset_y = _gdk_offset_y;
+	}
       break;
 
     case GDK_WINDOW_CHILD:
       dwStyle = WS_CHILDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
       break;
 
-    case GDK_WINDOW_DIALOG:
-      dwStyle = WS_OVERLAPPED | WS_MINIMIZEBOX | WS_SYSMENU | WS_CAPTION | WS_THICKFRAME | WS_CLIPCHILDREN;
-#if 0
-      dwExStyle |= WS_EX_TOPMOST; /* //HB: want this? */
-#endif
-      hparent = _gdk_root_window;
-      offset_x = _gdk_offset_x;
-      offset_y = _gdk_offset_y;
-      break;
-
     case GDK_WINDOW_TEMP:
-      dwStyle = WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
-      /* a temp window is not necessarily a top level window */
-      dwStyle |= (_gdk_parent_root == parent ? WS_POPUP : WS_CHILDWINDOW);
+      /* A temp window is not necessarily a top level window */
+      dwStyle = (_gdk_root == parent ? WS_POPUP : WS_CHILDWINDOW);
+      dwStyle |= WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
       dwExStyle |= WS_EX_TOOLWINDOW;
       offset_x = _gdk_offset_x;
       offset_y = _gdk_offset_y;
@@ -636,7 +684,7 @@ gdk_window_new_internal (GdkWindow     *parent,
   else
     title = get_default_title ();
   if (!title || !*title)
-    title = "GDK client window";
+    title = "";
 
   private->event_mask = GDK_STRUCTURE_MASK | attributes->event_mask;
       
@@ -647,20 +695,18 @@ gdk_window_new_internal (GdkWindow     *parent,
 
   mbtitle = g_locale_from_utf8 (title, -1, NULL, NULL, NULL);
   
-  {
-  HWND hwndNew =
-    CreateWindowEx (dwExStyle,
-		    MAKEINTRESOURCE(klass),
-		    mbtitle,
-		    dwStyle,
-		    ((attributes_mask & GDK_WA_X) ?
-		     impl->position_info.x - offset_x: CW_USEDEFAULT),
-		    impl->position_info.y - offset_y, 
-		    window_width, window_height,
-		    hparent,
-		    NULL,
-		    _gdk_app_hmodule,
-		    window);
+  hwndNew = CreateWindowEx (dwExStyle,
+			    MAKEINTRESOURCE(klass),
+			    mbtitle,
+			    dwStyle,
+			    ((attributes_mask & GDK_WA_X) ?
+			     impl->position_info.x - offset_x : CW_USEDEFAULT),
+			    impl->position_info.y - offset_y, 
+			    window_width, window_height,
+			    hparent,
+			    NULL,
+			    _gdk_app_hmodule,
+			    window);
   if (GDK_WINDOW_HWND (window) != hwndNew)
     {
       g_warning("gdk_window_new: gdk_event_translate::WM_CREATE (%p, %p) HWND mismatch.",
@@ -682,11 +728,11 @@ gdk_window_new_internal (GdkWindow     *parent,
 # endif
 
     }
-  }
+
   g_object_ref (window);
   gdk_win32_handle_table_insert (&GDK_WINDOW_HWND (window), window);
 
-  GDK_NOTE (MISC, g_print ("... \"%s\" %dx%d@+%d+%d %p = %p\n",
+  GDK_NOTE (MISC, g_print ("... \"%s\" %dx%d@%+d%+d %p = %p\n",
 			   mbtitle,
 			   window_width, window_height,
 			   ((attributes_mask & GDK_WA_X) ?
@@ -746,7 +792,7 @@ gdk_window_foreign_new_for_display (GdkDisplay      *display,
   
   private->parent = gdk_win32_handle_table_lookup ((GdkNativeWindow) parent);
   if (!private->parent || GDK_WINDOW_TYPE (private->parent) == GDK_WINDOW_FOREIGN)
-    private->parent = (GdkWindowObject *)_gdk_parent_root;
+    private->parent = (GdkWindowObject *)_gdk_root;
   
   private->parent->children = g_list_prepend (private->parent->children, window);
 
@@ -755,7 +801,7 @@ gdk_window_foreign_new_for_display (GdkDisplay      *display,
   point.x = rect.left;
   point.y = rect.right;
   ClientToScreen ((HWND) anid, &point);
-  if (parent != _gdk_root_window)
+  if (parent != GetDesktopWindow ())
     ScreenToClient (parent, &point);
   private->x = point.x;
   private->y = point.y;
@@ -781,6 +827,11 @@ gdk_window_foreign_new_for_display (GdkDisplay      *display,
   g_object_ref (window);
   gdk_win32_handle_table_insert (&GDK_WINDOW_HWND (window), window);
 
+  GDK_NOTE (MISC, g_print ("gdk_window_foreign_new_for_display: %p: %s@%+d%+d\n",
+			   (HWND) anid,
+			   _gdk_win32_drawable_description (window),
+			   private->x, private->y));
+
   return window;
 }
 
@@ -808,6 +859,8 @@ _gdk_windowing_window_destroy (GdkWindow *window,
 
   if (!recursing && !foreign_destroy)
     {
+      _gdk_win32_drawable_finish (private->impl);
+
       private->destroyed = TRUE;
       DestroyWindow (GDK_WINDOW_HWND (window));
     }
@@ -817,19 +870,13 @@ _gdk_windowing_window_destroy (GdkWindow *window,
 void
 _gdk_windowing_window_destroy_foreign (GdkWindow *window)
 {
-  /* It's somebody else's window, but in our hierarchy,
-   * so reparent it to the root window, and then call
-   * DestroyWindow() on it.
+  /* It's somebody else's window, but in our hierarchy, so reparent it
+   * to the desktop, and then try to destroy it.
    */
   gdk_window_hide (window);
   gdk_window_reparent (window, NULL, 0, 0);
   
-  /* Is this too drastic? Many (most?) applications
-   * quit if any window receives WM_QUIT I think.
-   * OTOH, I don't think foreign windows are much
-   * used, so the question is maybe academic.
-   */
-  PostMessage (GDK_WINDOW_HWND (window), WM_QUIT, 0, 0);
+  PostMessage (GDK_WINDOW_HWND (window), WM_CLOSE, 0, 0);
 }
 
 /* This function is called when the window really gone.
@@ -1025,8 +1072,7 @@ show_window_internal (GdkWindow *window,
       if (private->state & GDK_WINDOW_STATE_ABOVE)
         exstyle |= WS_EX_TOPMOST;
 
-      if (!SetWindowLong (GDK_WINDOW_HWND (window), GWL_EXSTYLE, exstyle))
-        WIN32_API_FAILED ("SetWindowLong");
+      API_CALL (SetWindowLong, (GDK_WINDOW_HWND (window), GWL_EXSTYLE, exstyle));
     }
 
   if (private->state & GDK_WINDOW_STATE_FULLSCREEN)
@@ -1043,10 +1089,11 @@ show_window_internal (GdkWindow *window,
   if (raise)
     {
       if (GDK_WINDOW_TYPE (window) == GDK_WINDOW_TEMP)
-        SetWindowPos (GDK_WINDOW_HWND (window), HWND_TOPMOST, 0, 0, 0, 0,
-	  	    SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
-      else if (GDK_WINDOW_TYPE (window) == GDK_WINDOW_TOPLEVEL
-	       || GDK_WINDOW_TYPE (window) == GDK_WINDOW_DIALOG)
+        SetWindowPos (GDK_WINDOW_HWND (window), HWND_TOPMOST,
+		      0, 0, 0, 0,
+		      SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
+      else if (GDK_WINDOW_TYPE (window) == GDK_WINDOW_TOPLEVEL ||
+	       GDK_WINDOW_TYPE (window) == GDK_WINDOW_DIALOG)
         SetForegroundWindow (GDK_WINDOW_HWND (window));
       else
         BringWindowToTop (GDK_WINDOW_HWND (window));
@@ -1098,7 +1145,8 @@ gdk_window_hide (GdkWindow *window)
   
   if (GetWindowLong (GDK_WINDOW_HWND (window), GWL_EXSTYLE) & WS_EX_TRANSPARENT)
     {
-      SetWindowPos (GDK_WINDOW_HWND (window), HWND_BOTTOM, 0, 0, 0, 0,
+      SetWindowPos (GDK_WINDOW_HWND (window), HWND_BOTTOM,
+		    0, 0, 0, 0,
 		    SWP_HIDEWINDOW | SWP_NOREDRAW | SWP_NOZORDER | SWP_NOMOVE | SWP_NOSIZE);
     }
   else
@@ -1139,7 +1187,7 @@ gdk_window_move (GdkWindow *window,
   if (GDK_WINDOW_DESTROYED (window))
     return;
 
-  GDK_NOTE (MISC, g_print ("gdk_window_move: %p: +%d+%d\n",
+  GDK_NOTE (MISC, g_print ("gdk_window_move: %p: %+d%+d\n",
 			   GDK_WINDOW_HWND (window), x, y));
       
   impl = GDK_WINDOW_IMPL_WIN32 (private->impl);
@@ -1147,7 +1195,11 @@ gdk_window_move (GdkWindow *window,
   if (private->state & GDK_WINDOW_STATE_FULLSCREEN)
     return;
 
-  if (GDK_WINDOW_TYPE (private) == GDK_WINDOW_CHILD)
+  /* Don't check GDK_WINDOW_TYPE (private) == GDK_WINDOW_CHILD.
+   * Foreign windows (another app's windows) might be children of our
+   * windows! Especially in the case of gtkplug/socket.
+   */ 
+  if (GetAncestor (GDK_WINDOW_HWND (window), GA_PARENT) != GetDesktopWindow ())
     _gdk_window_move_resize_child (window, x, y, impl->width, impl->height);
   else
     {
@@ -1157,6 +1209,11 @@ gdk_window_move (GdkWindow *window,
       
       adjust_for_gravity_hints (impl, &outer_rect, &x, &y);
 
+      GDK_NOTE (MISC, g_print ("... SetWindowPos(%p,NULL,%d,%d,0,0,"
+			       "NOACTIVATE|NOSIZE|NOZORDER)\n",
+			       GDK_WINDOW_HWND (window),
+			       x - _gdk_offset_x, y - _gdk_offset_y));
+
       API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window), NULL,
 			       x - _gdk_offset_x, y - _gdk_offset_y, 0, 0,
 			       SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER));
@@ -1190,13 +1247,19 @@ gdk_window_resize (GdkWindow *window,
   if (private->state & GDK_WINDOW_STATE_FULLSCREEN)
     return;
 
-  if (GDK_WINDOW_TYPE (private) == GDK_WINDOW_CHILD)
+  if (GetAncestor (GDK_WINDOW_HWND (window), GA_PARENT) != GetDesktopWindow ())
     _gdk_window_move_resize_child (window, private->x, private->y, width, height);
   else
     {
       RECT outer_rect;
       get_outer_rect (window, width, height, &outer_rect);
 
+      GDK_NOTE (MISC, g_print ("... SetWindowPos(%p,NULL,0,0,%ld,%ld,"
+			       "NOACTIVATE|NOMOVE|NOZORDER)\n",
+			       GDK_WINDOW_HWND (window),
+			       outer_rect.right - outer_rect.left,
+			       outer_rect.bottom - outer_rect.top));
+
       API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window), NULL,
 			       0, 0,
 			       outer_rect.right - outer_rect.left,
@@ -1232,11 +1295,11 @@ gdk_window_move_resize (GdkWindow *window,
   if (private->state & GDK_WINDOW_STATE_FULLSCREEN)
     return;
 
-  GDK_NOTE (MISC, g_print ("gdk_window_move_resize: %p: %dx%d@+%d+%d\n",
+  GDK_NOTE (MISC, g_print ("gdk_window_move_resize: %p: %dx%d@%+d%+d\n",
 			   GDK_WINDOW_HWND (window),
 			   width, height, x, y));
   
-  if (GDK_WINDOW_TYPE (private) == GDK_WINDOW_CHILD)
+  if (GetAncestor (GDK_WINDOW_HWND (window), GA_PARENT) != GetDesktopWindow ())
     _gdk_window_move_resize_child (window, x, y, width, height);
   else
     {
@@ -1246,6 +1309,13 @@ gdk_window_move_resize (GdkWindow *window,
 
       adjust_for_gravity_hints (impl, &outer_rect, &x, &y);
 
+      GDK_NOTE (MISC, g_print ("... SetWindowPos(%p,NULL,%d,%d,%ld,%ld,"
+			       "NOACTIVATE|NOZORDER)\n",
+			       GDK_WINDOW_HWND (window),
+			       x - _gdk_offset_x, y - _gdk_offset_y,
+			       outer_rect.right - outer_rect.left,
+			       outer_rect.bottom - outer_rect.top));
+			       
       API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window), NULL,
 			       x - _gdk_offset_x, y - _gdk_offset_y,
 			       outer_rect.right - outer_rect.left,
@@ -1264,11 +1334,13 @@ gdk_window_reparent (GdkWindow *window,
   GdkWindowObject *parent_private;
   GdkWindowObject *old_parent_private;
   GdkWindowImplWin32 *impl;
+  gboolean was_toplevel;
+  LONG style;
 
   g_return_if_fail (window != NULL);
   g_return_if_fail (GDK_IS_WINDOW (window));
   g_return_if_fail (new_parent == NULL || GDK_IS_WINDOW (new_parent));
-  g_return_if_fail (GDK_WINDOW_TYPE (window) != GDK_WINDOW_ROOT);
+  g_return_if_fail (window != _gdk_root);
 
   if (GDK_WINDOW_DESTROYED (window) ||
       (new_parent && GDK_WINDOW_DESTROYED (new_parent)))
@@ -1277,34 +1349,76 @@ gdk_window_reparent (GdkWindow *window,
     }
 
   if (!new_parent)
-    new_parent = _gdk_parent_root;
+    new_parent = _gdk_root;
 
   window_private = (GdkWindowObject*) window;
   old_parent_private = (GdkWindowObject *) window_private->parent;
   parent_private = (GdkWindowObject*) new_parent;
   impl = GDK_WINDOW_IMPL_WIN32 (window_private->impl);
 
-  if (!GDK_WINDOW_DESTROYED (window) && !GDK_WINDOW_DESTROYED (new_parent))
+  GDK_NOTE (MISC, g_print ("gdk_window_reparent: %p: %p\n",
+			   GDK_WINDOW_HWND (window),
+			   GDK_WINDOW_HWND (new_parent)));
+
+  style = GetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE);
+
+  was_toplevel = GetAncestor (GDK_WINDOW_HWND (window), GA_PARENT) == GetDesktopWindow ();
+  if (was_toplevel && new_parent != _gdk_root)
     {
-      GDK_NOTE (MISC, g_print ("gdk_window_reparent: %p: %p\n",
-			       GDK_WINDOW_HWND (window),
-			       GDK_WINDOW_HWND (new_parent)));
-
-      API_CALL (SetParent, (GDK_WINDOW_HWND (window),
-			    GDK_WINDOW_HWND (new_parent)));
-
-      API_CALL (MoveWindow, (GDK_WINDOW_HWND (window),
-			     x, y, impl->width, impl->height, TRUE));
+      /* Reparenting from top-level (child of desktop). Clear out
+       * decorations.
+       */
+      style &= ~(WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX);
+      style |= WS_CHILD;
+      SetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE, style);
     }
+  else if (new_parent == _gdk_root)
+    {
+      /* Reparenting to top-level. Add decorations. */
+      style &= ~(WS_CHILD);
+      style |= WS_OVERLAPPEDWINDOW;
+      SetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE, style);
+    }
+
+  API_CALL (SetParent, (GDK_WINDOW_HWND (window),
+			GDK_WINDOW_HWND (new_parent)));
+  
+  API_CALL (MoveWindow, (GDK_WINDOW_HWND (window),
+			 x, y, impl->width, impl->height, TRUE));
 
   /* From here on, we treat parents of type GDK_WINDOW_FOREIGN like
    * the root window
    */
   if (GDK_WINDOW_TYPE (new_parent) == GDK_WINDOW_FOREIGN)
-    new_parent = _gdk_parent_root;
+    new_parent = _gdk_root;
   
   window_private->parent = (GdkWindowObject *)new_parent;
 
+  /* Switch the window type as appropriate */
+
+  switch (GDK_WINDOW_TYPE (new_parent))
+    {
+    case GDK_WINDOW_ROOT:
+      if (impl->toplevel_window_type != -1)
+	GDK_WINDOW_TYPE (window) = impl->toplevel_window_type;
+      else if (GDK_WINDOW_TYPE (window) == GDK_WINDOW_CHILD)
+	GDK_WINDOW_TYPE (window) = GDK_WINDOW_TOPLEVEL;
+      break;
+
+    case GDK_WINDOW_TOPLEVEL:
+    case GDK_WINDOW_CHILD:
+    case GDK_WINDOW_DIALOG:
+    case GDK_WINDOW_TEMP:
+      if (WINDOW_IS_TOPLEVEL (window))
+	{
+	  /* Save the original window type so we can restore it if the
+	   * window is reparented back to be a toplevel.
+	   */
+	  impl->toplevel_window_type = GDK_WINDOW_TYPE (window);
+	  GDK_WINDOW_TYPE (window) = GDK_WINDOW_CHILD;
+	}
+    }
+
   if (old_parent_private)
     old_parent_private->children =
       g_list_remove (old_parent_private->children, window);
@@ -1336,7 +1450,7 @@ _gdk_windowing_window_clear_area (GdkWindow *window,
       if (height == 0)
 	height = impl->height - y;
       GDK_NOTE (MISC, g_print ("_gdk_windowing_window_clear_area: %p: "
-			       "%dx%d@+%d+%d\n",
+			       "%dx%d@%+d%+d\n",
 			       GDK_WINDOW_HWND (window),
 			       width, height, x, y));
       hdc = GetDC (GDK_WINDOW_HWND (window));
@@ -1361,7 +1475,7 @@ _gdk_windowing_window_clear_area_e (GdkWindow *window,
       RECT rect;
 
       GDK_NOTE (MISC, g_print ("_gdk_windowing_window_clear_area_e: %p: "
-			       "%dx%d@+%d+%d\n",
+			       "%dx%d@%+d%+d\n",
 			       GDK_WINDOW_HWND (window),
 			       width, height, x, y));
 
@@ -1397,10 +1511,14 @@ gdk_window_lower (GdkWindow *window)
   
   if (!GDK_WINDOW_DESTROYED (window))
     {
-      GDK_NOTE (MISC, g_print ("gdk_window_lower: %p\n",
+      GDK_NOTE (MISC, g_print ("gdk_window_lower: %p\n"
+			       "... SetWindowPos(%p,HWND_BOTTOM,0,0,0,0,"
+			       "NOACTIVATE|NOMOVE|NOSIZE)\n",
+			       GDK_WINDOW_HWND (window),
 			       GDK_WINDOW_HWND (window)));
 
-      API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window), HWND_BOTTOM, 0, 0, 0, 0,
+      API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window), HWND_BOTTOM,
+			       0, 0, 0, 0,
 			       SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE));
     }
 }
@@ -1427,7 +1545,7 @@ gdk_window_set_hints (GdkWindow *window,
   
   impl = GDK_WINDOW_IMPL_WIN32 (GDK_WINDOW_OBJECT (window)->impl);
 
-  GDK_NOTE (MISC, g_print ("gdk_window_set_hints: %p: %dx%d..%dx%d @+%d+%d\n",
+  GDK_NOTE (MISC, g_print ("gdk_window_set_hints: %p: %dx%d..%dx%d @%+d%+d\n",
 			   GDK_WINDOW_HWND (window),
 			   min_width, min_height, max_width, max_height,
 			   x, y));
@@ -1452,6 +1570,30 @@ gdk_window_set_hints (GdkWindow *window,
     }
 }
 
+void
+gdk_window_set_urgency_hint (GdkWindow *window,
+			     gboolean   urgent)
+{
+  FLASHWINFO flashwinfo;
+
+  g_return_if_fail (GDK_IS_WINDOW (window));
+  g_return_if_fail (GDK_WINDOW_TYPE (window) != GDK_WINDOW_CHILD);
+  
+  if (GDK_WINDOW_DESTROYED (window))
+    return;
+
+  flashwinfo.cbSize = sizeof (flashwinfo);
+  flashwinfo.hwnd = GDK_WINDOW_HWND (window);
+  if (urgent)
+    flashwinfo.dwFlags = FLASHW_ALL | FLASHW_TIMER;
+  else
+    flashwinfo.dwFlags = FLASHW_STOP;
+  flashwinfo.uCount = 0;
+  flashwinfo.dwTimeout = 0;
+
+  FlashWindowEx (&flashwinfo);
+}
+
 void 
 gdk_window_set_geometry_hints (GdkWindow      *window,
 			       GdkGeometry    *geometry,
@@ -1563,7 +1705,7 @@ gdk_window_set_geometry_hints (GdkWindow      *window,
 	    size_hints.rcNormalPosition.left + geometry->base_width;
 	  size_hints.rcNormalPosition.bottom =
 	    size_hints.rcNormalPosition.top + geometry->base_height;
-	  GDK_NOTE (MISC, g_print ("...setting: rcNormal: (%ld,%ld)--(%ld,%ld)\n",
+	  GDK_NOTE (MISC, g_print ("... setting: rcNormal: (%ld,%ld)--(%ld,%ld)\n",
 				   size_hints.rcNormalPosition.left,
 				   size_hints.rcNormalPosition.top,
 				   size_hints.rcNormalPosition.right,
@@ -1595,12 +1737,12 @@ void
 gdk_window_set_title (GdkWindow   *window,
 		      const gchar *title)
 {
-  char *mbtitle;
-
-  g_return_if_fail (window != NULL);
   g_return_if_fail (GDK_IS_WINDOW (window));
   g_return_if_fail (title != NULL);
 
+  if (GDK_WINDOW_DESTROYED (window))
+    return;
+
   /* Empty window titles not allowed, so set it to just a period. */
   if (!title[0])
     title = ".";
@@ -1608,14 +1750,17 @@ gdk_window_set_title (GdkWindow   *window,
   GDK_NOTE (MISC, g_print ("gdk_window_set_title: %p: %s\n",
 			   GDK_WINDOW_HWND (window), title));
   
-  if (!GDK_WINDOW_DESTROYED (window))
+  if (G_WIN32_HAVE_WIDECHAR_API ())
     {
-      /* As the title is in UTF-8 we must translate it
-       * to the system codepage.
-       */
-      mbtitle = g_locale_from_utf8 (title, -1, NULL, NULL, NULL);
-      API_CALL (SetWindowText, (GDK_WINDOW_HWND (window), mbtitle));
-      g_free (mbtitle);
+      wchar_t *wtitle = g_utf8_to_utf16 (title, -1, NULL, NULL, NULL);
+      API_CALL (SetWindowTextW, (GDK_WINDOW_HWND (window), wtitle));
+      g_free (wtitle);
+    }
+  else
+    {
+      char *cptitle = g_locale_from_utf8 (title, -1, NULL, NULL, NULL);
+      API_CALL (SetWindowTextA, (GDK_WINDOW_HWND (window), cptitle));
+      g_free (cptitle);
     }
 }
 
@@ -1650,7 +1795,7 @@ gdk_window_set_transient_for (GdkWindow *window,
 
   if (((GdkWindowObject *) window)->window_type == GDK_WINDOW_CHILD)
     {
-      GDK_NOTE (MISC, g_print ("...a child window!\n"));
+      GDK_NOTE (MISC, g_print ("... a child window!\n"));
       return;
     }
   
@@ -1704,6 +1849,12 @@ gdk_window_set_back_pixmap (GdkWindow *window,
   g_return_if_fail (pixmap == NULL || !parent_relative);
   g_return_if_fail (pixmap == NULL || gdk_drawable_get_depth (window) == gdk_drawable_get_depth (pixmap));
   
+  if (!gdk_drawable_get_colormap (pixmap))
+    {
+      g_warning ("gdk_window_set_back_pixmap(): pixmap must have a colormap");
+      return;
+    }
+  
   if (private->bg_pixmap &&
       private->bg_pixmap != GDK_PARENT_RELATIVE_BG &&
       private->bg_pixmap != GDK_NO_BG)
@@ -1774,7 +1925,7 @@ gdk_window_set_cursor (GdkWindow *window,
        */
       if ((impl->hcursor = CopyCursor (hcursor)) == NULL)
 	WIN32_API_FAILED ("CopyCursor");
-      GDK_NOTE (MISC, g_print ("...CopyCursor (%p) = %p\n",
+      GDK_NOTE (MISC, g_print ("... CopyCursor (%p) = %p\n",
 			       hcursor, impl->hcursor));
     }
 
@@ -1831,8 +1982,7 @@ gdk_window_set_cursor (GdkWindow *window,
           SetCursor (hcursor);
         }
 
-      GDK_NOTE (MISC, g_print ("...DestroyCursor (%p)\n",
-			       hprevcursor));
+      GDK_NOTE (MISC, g_print ("... DestroyCursor (%p)\n", hprevcursor));
       
       API_CALL (DestroyCursor, (hprevcursor));
     }
@@ -1849,7 +1999,7 @@ gdk_window_get_geometry (GdkWindow *window,
   g_return_if_fail (window == NULL || GDK_IS_WINDOW (window));
   
   if (!window)
-    window = _gdk_parent_root;
+    window = _gdk_root;
   
   if (!GDK_WINDOW_DESTROYED (window))
     {
@@ -1857,7 +2007,7 @@ gdk_window_get_geometry (GdkWindow *window,
 
       API_CALL (GetClientRect, (GDK_WINDOW_HWND (window), &rect));
 
-      if (window != _gdk_parent_root)
+      if (window != _gdk_root)
 	{
 	  POINT pt;
 	  GdkWindow *parent = gdk_window_get_parent (window);
@@ -1876,7 +2026,7 @@ gdk_window_get_geometry (GdkWindow *window,
 	  rect.right = pt.x;
 	  rect.bottom = pt.y;
 
-	  if (parent == _gdk_parent_root)
+	  if (parent == _gdk_root)
 	    {
 	      rect.left += _gdk_offset_x;
 	      rect.top += _gdk_offset_y;
@@ -1896,7 +2046,7 @@ gdk_window_get_geometry (GdkWindow *window,
       if (depth)
 	*depth = gdk_drawable_get_visual (window)->depth;
 
-      GDK_NOTE (MISC, g_print ("gdk_window_get_geometry: %p: %ldx%ldx%d@+%ld+%ld\n",
+      GDK_NOTE (MISC, g_print ("gdk_window_get_geometry: %p: %ldx%ldx%d@%+ld%+ld\n",
 			       GDK_WINDOW_HWND (window),
 			       rect.right - rect.left, rect.bottom - rect.top,
 			       gdk_drawable_get_visual (window)->depth,
@@ -1934,7 +2084,7 @@ gdk_window_get_origin (GdkWindow *window,
   if (y)
     *y = ty + _gdk_offset_y;
 
-  GDK_NOTE (MISC, g_print ("gdk_window_get_origin: %p: +%d+%d\n",
+  GDK_NOTE (MISC, g_print ("gdk_window_get_origin: %p: %+d%+d\n",
 			   GDK_WINDOW_HWND (window),
 			   tx, ty));
   return return_val;
@@ -1965,7 +2115,7 @@ gdk_window_get_root_origin (GdkWindow *window,
   if (y)
     *y = rect.y;
 
-  GDK_NOTE (MISC, g_print ("gdk_window_get_root_origin: %p: +%d+%d\n",
+  GDK_NOTE (MISC, g_print ("gdk_window_get_root_origin: %p: %+d%+d\n",
 			   GDK_WINDOW_HWND (window), rect.x, rect.y));
 }
 
@@ -2004,7 +2154,7 @@ gdk_window_get_frame_extents (GdkWindow    *window,
   rect->width = r.right - r.left;
   rect->height = r.bottom - r.top;
 
-  GDK_NOTE (MISC, g_print ("gdk_window_get_frame_extents: %p: %ldx%ld@+%ld+%ld\n",
+  GDK_NOTE (MISC, g_print ("gdk_window_get_frame_extents: %p: %ldx%ld@%+ld%+ld\n",
 			   GDK_WINDOW_HWND (window),
 			   r.right - r.left, r.bottom - r.top,
 			   r.left, r.top));
@@ -2032,7 +2182,7 @@ _gdk_windowing_window_get_pointer (GdkDisplay      *display,
   *x = point.x;
   *y = point.y;
 
-  if (window == _gdk_parent_root)
+  if (window == _gdk_root)
     {
       *x += _gdk_offset_x;
       *y += _gdk_offset_y;
@@ -2095,6 +2245,15 @@ _gdk_windowing_get_pointer (GdkDisplay       *display,
   _gdk_windowing_window_get_pointer (display, root_window, x, y, mask);
 }
 
+void
+gdk_display_warp_pointer (GdkDisplay *display,
+			  GdkScreen  *screen,
+			  gint        x,
+			  gint        y)
+{
+  SetCursorPos (x, y);
+}
+
 GdkWindow*
 _gdk_windowing_window_at_pointer (GdkDisplay *display,
 				  gint       *win_x,
@@ -2111,7 +2270,7 @@ _gdk_windowing_window_at_pointer (GdkDisplay *display,
 
   if (hwnd == NULL)
     {
-      window = _gdk_parent_root;
+      window = _gdk_root;
       *win_x = pointc.x + _gdk_offset_x;
       *win_y = pointc.y + _gdk_offset_y;
       return window;
@@ -2134,7 +2293,7 @@ _gdk_windowing_window_at_pointer (GdkDisplay *display,
       *win_y = point.y - rect.top;
     }
 
-  GDK_NOTE (MISC, g_print ("_gdk_windowing_window_at_pointer: +%d+%d %p%s\n",
+  GDK_NOTE (MISC, g_print ("_gdk_windowing_window_at_pointer: %+d%+d %p%s\n",
 			   *win_x, *win_y,
 			   hwnd,
 			   (window == NULL ? " NULL" : "")));
@@ -2434,7 +2593,8 @@ gdk_window_set_decorations (GdkWindow      *window,
     style &= ~settable_bits, style |= bits;
   
   SetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE, style);
-  SetWindowPos (GDK_WINDOW_HWND (window), NULL, 0, 0, 0, 0,
+  SetWindowPos (GDK_WINDOW_HWND (window), NULL,
+		0, 0, 0, 0,
 		SWP_FRAMECHANGED | SWP_NOACTIVATE | SWP_NOMOVE |
 		SWP_NOREPOSITION | SWP_NOSIZE | SWP_NOZORDER);
 }
@@ -2503,7 +2663,8 @@ gdk_window_set_functions (GdkWindow    *window,
     style &= ~settable_bits, style |= bits;
   
   SetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE, style);
-  SetWindowPos (GDK_WINDOW_HWND (window), NULL, 0, 0, 0, 0,
+  SetWindowPos (GDK_WINDOW_HWND (window), NULL,
+		0, 0, 0, 0,
 		SWP_FRAMECHANGED | SWP_NOACTIVATE | SWP_NOMOVE |
 		SWP_NOREPOSITION | SWP_NOSIZE | SWP_NOZORDER);
 }
@@ -2607,8 +2768,6 @@ gboolean
 gdk_window_set_static_gravities (GdkWindow *window,
 				 gboolean   use_static)
 {
-  GdkWindowObject *private = (GdkWindowObject *)window;
-  
   g_return_val_if_fail (window != NULL, FALSE);
   g_return_val_if_fail (GDK_IS_WINDOW (window), FALSE);
 
@@ -2868,10 +3027,10 @@ gdk_window_fullscreen (GdkWindow *window)
 
       SetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE, 
                      (fi->style & ~WS_OVERLAPPEDWINDOW) | WS_POPUP);
-      if (!SetWindowPos (GDK_WINDOW_HWND (window), HWND_TOP /*not MOST, taskswitch!*/,
-                         0, 0, width, height,
-                         SWP_NOCOPYBITS | SWP_SHOWWINDOW))
-        WIN32_API_FAILED ("SetWindowPos");
+
+      API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window), HWND_TOP,
+			       0, 0, width, height,
+			       SWP_NOCOPYBITS | SWP_SHOWWINDOW));
 
       gdk_synthesize_window_state (window, 0, GDK_WINDOW_STATE_FULLSCREEN);
     }
@@ -2892,11 +3051,11 @@ gdk_window_unfullscreen (GdkWindow *window)
 
       impl->hint_flags = fi->hint_flags;
       SetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE, fi->style);
-      if (!SetWindowPos (GDK_WINDOW_HWND (window), HWND_NOTOPMOST,
-                         fi->r.left, fi->r.top, fi->r.right - fi->r.left, fi->r.bottom - fi->r.top,
-                         SWP_NOCOPYBITS | SWP_SHOWWINDOW))
-         WIN32_API_FAILED ("SetWindowPos");
-
+      API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window), HWND_NOTOPMOST,
+			       fi->r.left, fi->r.top,
+			       fi->r.right - fi->r.left, fi->r.bottom - fi->r.top,
+			       SWP_NOCOPYBITS | SWP_SHOWWINDOW));
+      
       g_object_set_data (G_OBJECT (window), "fullscreen-info", NULL);
       g_free (fi);
 
@@ -2914,9 +3073,10 @@ gdk_window_set_keep_above (GdkWindow *window, gboolean setting)
 
   if (GDK_WINDOW_IS_MAPPED (window))
     {
-      if (!SetWindowPos(GDK_WINDOW_HWND (window), setting ? HWND_TOPMOST : HWND_NOTOPMOST,
-                        0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE))
-         WIN32_API_FAILED ("SetWindowPos");
+      API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window),
+			       setting ? HWND_TOPMOST : HWND_NOTOPMOST,
+			       0, 0, 0, 0,
+			       SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE));
     }
   else
     gdk_synthesize_window_state (window,
@@ -2934,9 +3094,10 @@ gdk_window_set_keep_below (GdkWindow *window, gboolean setting)
 
   if (GDK_WINDOW_IS_MAPPED (window))
     {
-      if (!SetWindowPos(GDK_WINDOW_HWND (window), setting ? HWND_BOTTOM : HWND_NOTOPMOST,
-                        0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE))
-         WIN32_API_FAILED ("SetWindowPos");
+      API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window),
+			       setting ? HWND_BOTTOM : HWND_NOTOPMOST,
+			       0, 0, 0, 0,
+			       SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE));
     }
   else
     gdk_synthesize_window_state (window,
@@ -2983,7 +3144,8 @@ gdk_window_set_modal_hint (GdkWindow *window,
   if (GDK_WINDOW_IS_MAPPED (window))
     API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window), 
 			     modal ? HWND_TOPMOST : HWND_NOTOPMOST,
-			     0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE));
+			     0, 0, 0, 0,
+			     SWP_NOMOVE | SWP_NOSIZE));
 }
 
 void
@@ -3016,7 +3178,8 @@ gdk_window_set_skip_taskbar_hint (GdkWindow *window,
 #if 0 /* Should we also turn off the minimize and maximize buttons? */
       SetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE,
 		     GetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE) & ~(WS_MINIMIZEBOX|WS_MAXIMIZEBOX|WS_SYSMENU));
-      SetWindowPos (GDK_WINDOW_HWND (window), NULL, 0, 0, 0, 0,
+      SetWindowPos (GDK_WINDOW_HWND (window), NULL,
+		    0, 0, 0, 0,
 		    SWP_FRAMECHANGED | SWP_NOACTIVATE | SWP_NOMOVE |
 		    SWP_NOREPOSITION | SWP_NOSIZE | SWP_NOZORDER);
 #endif
